// -----------------------------------------------------------------------------
// (C) Bibix
// The content below includes confidential, proprietary information of
// Bibix. All use, disclosure, and/or reproduction is prohibited
// unless authorized in writing. All rights reserved.
// -----------------------------------------------------------------------------
// File         : arx_numeric.h
// Description  : Header file for Arx numeric functions in C
// Author       : Klaas Hofstra, University of Twente
// Creation date: June 29, 2010 (Header added by SHG)
// -----------------------------------------------------------------------------
// $Rev: 47 $
// $Author: sabih $
// $Date: 2011-02-03 01:25:20 +0100 (Thu, 03 Feb 2011) $
// $Log$
// -----------------------------------------------------------------------------

#ifndef ARX_NUMERIC_H
#define ARX_NUMERIC_H

#include "arx_types.h"

// 2006-05-01
// - Fix for arx_slice_lhs and ars_bitindex_lhs. Now using sign extension.
// 2006-07-11
// - Added windows support.
// 2006-07-18
// - Converted arx_slice macro into inline functions.
// 2006-07-30
// - replaced types with i32,u32,i64,u64
// 2006-07-31
// - added arx_monitor

class arx_monitor {
public:

	arx_monitor() {
		reset();
	}

	static arx_monitor& instance() {
		static arx_monitor _instance;
		return _instance;
	}

	void reset() {
		sat_count = 0;
	}
	
	u32 get_sat_count() {
		return sat_count;
	}
	
	void inc_sat_count() {
		sat_count++;
	}
	
private:
	u32 sat_count;
};

template <typename T>
inline T arx_modulo(T in1, T in2)
{
	return in1 % in2;
}


/*
	Min Max
*/

template <typename T>
inline T arx_min(T in1, T in2)
{
	return in1 < in2 ? in1 : in2;
}

template <typename T>
inline T arx_max(T in1, T in2)
{
	return in1 > in2 ? in1 : in2;
}


/*
    Slice
*/
/*
#define arx_slice(in, left, right, out_type) \
    ( (out_type)((in) >> (right)) & ((1 << ((left) - (right) + 1)) - 1) )
*/

template <typename T>
inline u32 arx_slice32(const T in, int left, int right)
{
	T tmp = in >> right;
	T mask = (T(1) << (left-right+1)) - 1;
	return (tmp & mask);
}

inline u64 arx_slice64(u64 in, int left, int right)
{
	u64 tmp = in >> right;
	u64 mask = (u64(1) << (left-right+1)) - 1;
	return (tmp & mask);
}

template <typename T>
inline T arx_slice_lhs(const T lhs, const T rhs, int left, int right, int extend)
{
	T rhs_mask = (T(1) << (left-right+1)) - 1;
	T lhs_mask = ~(rhs_mask << right);
	T out = ( (lhs & lhs_mask) | ((rhs & rhs_mask) << right) );
	return ((out << extend) >> extend);	
}

/*
    Bitselect
*/

template <typename T>
inline u32 arx_bitselect(T in, int index)
{
	return ((in >> index) & 1);
}

template <typename T>
inline T arx_bitselect_lhs(T lhs, T rhs, int index, int extend)
{
	T mask = ~(T(1) << index);
	T out = ( (lhs & mask) | ((rhs & 1) << index) );
	return ((out << extend) >> extend);
}

/*
    Extend and mask functions
*/

template <typename T>
inline T arx_lsb_extend(T in, int nr_bits)
{
	return (in << nr_bits);
}

template <typename T>
inline T arx_sign_extend(T in, int nr_bits)
{
	return arx_wrap(in, nr_bits);
}

template <typename T>
inline T arx_mask_sign(T in, int nr_bits)
{
	T mask = (T(1) << nr_bits)-1;
	return (in & mask);
}


/*
    Overflow
*/

template <typename T>
inline T arx_wrap(T in, int nr_bits)
{
	return ((in << nr_bits) >> nr_bits);
}

/*
    Saturate
*/

template <typename T>
inline T arx_saturate_s(T in, int nr_bits, bool monitor)
{
	T max = (T(1) << (nr_bits-1)) - 1;
	T min = -(T(1) << (nr_bits-1));
	if (in > max) {
		if (monitor) {
			arx_monitor::instance().inc_sat_count();
		}
		return max;
	}
	else if (in < min) {
		if (monitor) {
			arx_monitor::instance().inc_sat_count();
		}
		return min;
	}
	return in;
}

template <typename T>
inline T arx_saturate_u(T in, int nr_bits, bool monitor)
{
	T max = (T(1) << nr_bits) - 1;
	if (in > max) {
		if (monitor) {
			arx_monitor::instance().inc_sat_count();
		}
		return max;
	}
	return in;
}

template <typename T>
inline T arx_saturate(T in, int nr_bits, bool monitor)
{
	return arx_saturate_s(in, nr_bits, monitor);
}

template <>
inline u32 arx_saturate(u32 in, int nr_bits, bool monitor)
{
	return arx_saturate_u(in, nr_bits, monitor);
}

template <>
inline u64 arx_saturate(u64 in, int nr_bits, bool monitor)
{
	return arx_saturate_u(in, nr_bits, monitor);
}

/*
    Saturate symmetrical
*/

template <typename T>
inline T arx_saturate_sym(T in, int nr_bits, bool monitor)
{
	T max = (T(1) << (nr_bits-1)) - 1;
	T min = -(T(1) << (nr_bits-1)) + 1;
	if (in > max) {
		if (monitor) {
			arx_monitor::instance().inc_sat_count();
		}
		return max;
	}
	else if (in < min) {
		if (monitor) {
			arx_monitor::instance().inc_sat_count();
		}
		return min;
	}
	return in;
}

/*
    Quantization
*/

template <typename T>
inline T arx_truncate(T in, int nr_bits)
{
	return (in >> nr_bits);
}

template <typename T>
inline T arx_round(T in, int nr_bits)
{
	T carry = (T(1) << (nr_bits - 1));
	return ((in + carry) >> nr_bits);
}


template <typename T>
inline T arx_round_zero_s(T in, int nr_bits)
{
	T half = in & ((T(1) << (nr_bits - 1)) - 1);
	if (half == 0) {
		if (in < 0) {
			return arx_round(in, nr_bits);
		}
		else {
			return arx_truncate(in, nr_bits);
		}
	}
	else {
		return arx_round(in, nr_bits);
	}
}

template <typename T>
inline T arx_round_zero_u(T in, int nr_bits)
{
	T half = in & ((T(1) << (nr_bits - 1)) - 1);
	if (half == 0) {
		return arx_truncate(in, nr_bits);
	}
	else {
		return arx_round(in, nr_bits);
	}
}

template <typename T>
inline T arx_round_zero(T in, int nr_bits)
{
	return arx_round_zero_u(in, nr_bits);
}

template <>
inline i32 arx_round_zero(i32 in, int nr_bits)
{
	return arx_round_zero_s(in, nr_bits);
}

template <>
inline i64 arx_round_zero(i64 in, int nr_bits)
{
	return arx_round_zero_s(in, nr_bits);
}

template <typename T>
inline T arx_round_inf_s(T in, int nr_bits)
{
	T half = in & ((T(1) << (nr_bits - 1)) - 1);
	if (half == 0) {
		if (in > 0) {
			return arx_round(in, nr_bits);
		}
		else {
			return arx_truncate(in, nr_bits);
		}
	}
	else {
		return arx_round(in, nr_bits);
	}
}

template <typename T>
inline T arx_round_inf(T in, int nr_bits)
{
	return arx_round(in, nr_bits);
}

template <>
inline i32 arx_round_inf(i32 in, int nr_bits)
{
	return arx_round_inf_s(in, nr_bits);
}

template <>
inline i64 arx_round_inf(i64 in, int nr_bits)
{
	return arx_round_inf_s(in, nr_bits);
}

#endif

